#nebulaexport.py, a nebula mesh/skeleton/animation exporter for the Nebula Device
#
# Portions of this code are Copyright (C) 2003 Jean-Baptiste LAMY -- jiba@tuxfamily.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# This script converts a mesh,skeleton, and action ipos into appropriate
# nebula1 files: a nebula .n3d mesh file, a .anim animation file, and
# a .py python script which initalizes the character node joints.

from Blender import NMesh, Ipo, Armature, Object, Material, Scene, Types, BGL, Draw

import sys, operator, string, math, os

# this is the directory into which the character files will be put
characterdir = "/home/delire/downloads/fijuu/nebula/data/meshes/"

# the default file name: produces files named character.n3d, character.anim, and character.py
defaultfilename = "character"

# comment string to stick at the top of the character.py file
scriptheader = """
# Character initialization script for exported blender character
"""

#===========================================================================================
# Math stuff - this is all shamelessy copied from Jean-Baptiste's exporter code - GJH

def quaternion2matrix(q):
  xx = q[0] * q[0]
  yy = q[1] * q[1]
  zz = q[2] * q[2]
  xy = q[0] * q[1]
  xz = q[0] * q[2]
  yz = q[1] * q[2]
  wx = q[3] * q[0]
  wy = q[3] * q[1]
  wz = q[3] * q[2]
  return [[1.0 - 2.0 * (yy + zz),       2.0 * (xy + wz),       2.0 * (xz - wy), 0.0],
          [      2.0 * (xy - wz), 1.0 - 2.0 * (xx + zz),       2.0 * (yz + wx), 0.0],
          [      2.0 * (xz + wy),       2.0 * (yz - wx), 1.0 - 2.0 * (xx + yy), 0.0],
          [0.0                  , 0.0                  , 0.0                  , 1.0]]

def matrix2quaternion(m):
  s = math.sqrt(abs(m[0][0] + m[1][1] + m[2][2] + m[3][3]))
  if s == 0.0:
    x = abs(m[2][1] - m[1][2])
    y = abs(m[0][2] - m[2][0])
    z = abs(m[1][0] - m[0][1])
    if   (x >= y) and (x >= z): return 1.0, 0.0, 0.0, 0.0
    elif (y >= x) and (y >= z): return 0.0, 1.0, 0.0, 0.0
    else:                       return 0.0, 0.0, 1.0, 0.0
  return quaternion_normalize([
    -(m[2][1] - m[1][2]) / (2.0 * s),
    -(m[0][2] - m[2][0]) / (2.0 * s),
    -(m[1][0] - m[0][1]) / (2.0 * s),
    0.5 * s,
    ])

def quaternion_normalize(q):
  l = math.sqrt(q[0] * q[0] + q[1] * q[1] + q[2] * q[2] + q[3] * q[3])
  return q[0] / l, q[1] / l, q[2] / l, q[3] / l

def quaternion_multiply(q1, q2):
  r = [
    q2[3] * q1[0] + q2[0] * q1[3] + q2[1] * q1[2] - q2[2] * q1[1],
    q2[3] * q1[1] + q2[1] * q1[3] + q2[2] * q1[0] - q2[0] * q1[2],
    q2[3] * q1[2] + q2[2] * q1[3] + q2[0] * q1[1] - q2[1] * q1[0],
    q2[3] * q1[3] - q2[0] * q1[0] - q2[1] * q1[1] - q2[2] * q1[2],
    ]
  d = math.sqrt(r[0] * r[0] + r[1] * r[1] + r[2] * r[2] + r[3] * r[3])
  r[0] /= d
  r[1] /= d
  r[2] /= d
  r[3] /= d
  return r

def matrix_translate(m, v):
  m[3][0] += v[0]
  m[3][1] += v[1]
  m[3][2] += v[2]
  return m

def matrix_multiply(b, a):
  return [ [
    a[0][0] * b[0][0] + a[0][1] * b[1][0] + a[0][2] * b[2][0],
    a[0][0] * b[0][1] + a[0][1] * b[1][1] + a[0][2] * b[2][1],
    a[0][0] * b[0][2] + a[0][1] * b[1][2] + a[0][2] * b[2][2],
    0.0,
    ], [
    a[1][0] * b[0][0] + a[1][1] * b[1][0] + a[1][2] * b[2][0],
    a[1][0] * b[0][1] + a[1][1] * b[1][1] + a[1][2] * b[2][1],
    a[1][0] * b[0][2] + a[1][1] * b[1][2] + a[1][2] * b[2][2],
    0.0,
    ], [
    a[2][0] * b[0][0] + a[2][1] * b[1][0] + a[2][2] * b[2][0],
    a[2][0] * b[0][1] + a[2][1] * b[1][1] + a[2][2] * b[2][1],
    a[2][0] * b[0][2] + a[2][1] * b[1][2] + a[2][2] * b[2][2],
     0.0,
    ], [
    a[3][0] * b[0][0] + a[3][1] * b[1][0] + a[3][2] * b[2][0] + b[3][0],
    a[3][0] * b[0][1] + a[3][1] * b[1][1] + a[3][2] * b[2][1] + b[3][1],
    a[3][0] * b[0][2] + a[3][1] * b[1][2] + a[3][2] * b[2][2] + b[3][2],
    1.0,
    ] ]

def matrix_invert(m):
  det = (m[0][0] * (m[1][1] * m[2][2] - m[2][1] * m[1][2])
       - m[1][0] * (m[0][1] * m[2][2] - m[2][1] * m[0][2])
       + m[2][0] * (m[0][1] * m[1][2] - m[1][1] * m[0][2]))
  if det == 0.0: return None
  det = 1.0 / det
  r = [ [
      det * (m[1][1] * m[2][2] - m[2][1] * m[1][2]),
    - det * (m[0][1] * m[2][2] - m[2][1] * m[0][2]),
      det * (m[0][1] * m[1][2] - m[1][1] * m[0][2]),
      0.0,
    ], [
    - det * (m[1][0] * m[2][2] - m[2][0] * m[1][2]),
      det * (m[0][0] * m[2][2] - m[2][0] * m[0][2]),
    - det * (m[0][0] * m[1][2] - m[1][0] * m[0][2]),
      0.0
    ], [
      det * (m[1][0] * m[2][1] - m[2][0] * m[1][1]),
    - det * (m[0][0] * m[2][1] - m[2][0] * m[0][1]),
      det * (m[0][0] * m[1][1] - m[1][0] * m[0][1]),
      0.0,
    ] ]
  r.append([
    -(m[3][0] * r[0][0] + m[3][1] * r[1][0] + m[3][2] * r[2][0]),
    -(m[3][0] * r[0][1] + m[3][1] * r[1][1] + m[3][2] * r[2][1]),
    -(m[3][0] * r[0][2] + m[3][1] * r[1][2] + m[3][2] * r[2][2]),
    1.0,
    ])
  return r

def matrix_rotate_x(angle):
  cos = math.cos(angle)
  sin = math.sin(angle)
  return [
    [1.0,  0.0, 0.0, 0.0],
    [0.0,  cos, sin, 0.0],
    [0.0, -sin, cos, 0.0],
    [0.0,  0.0, 0.0, 1.0],
    ]

def matrix_rotate_y(angle):
  cos = math.cos(angle)
  sin = math.sin(angle)
  return [
    [cos, 0.0, -sin, 0.0],
    [0.0, 1.0,  0.0, 0.0],
    [sin, 0.0,  cos, 0.0],
    [0.0, 0.0,  0.0, 1.0],
    ]

def matrix_rotate_z(angle):
  cos = math.cos(angle)
  sin = math.sin(angle)
  return [
    [ cos, sin, 0.0, 0.0],
    [-sin, cos, 0.0, 0.0],
    [ 0.0, 0.0, 1.0, 0.0],
    [ 0.0, 0.0, 0.0, 1.0],
    ]

def matrix_rotate(axis, angle):
  vx  = axis[0]
  vy  = axis[1]
  vz  = axis[2]
  vx2 = vx * vx
  vy2 = vy * vy
  vz2 = vz * vz
  cos = math.cos(angle)
  sin = math.sin(angle)
  co1 = 1.0 - cos
  return [
    [vx2 * co1 + cos,          vx * vy * co1 + vz * sin, vz * vx * co1 - vy * sin, 0.0],
    [vx * vy * co1 - vz * sin, vy2 * co1 + cos,          vy * vz * co1 + vx * sin, 0.0],
    [vz * vx * co1 + vy * sin, vy * vz * co1 - vx * sin, vz2 * co1 + cos,          0.0],
    [0.0, 0.0, 0.0, 1.0],
    ]
  
def point_by_matrix(p, m):
  return [p[0] * m[0][0] + p[1] * m[1][0] + p[2] * m[2][0] + m[3][0],
          p[0] * m[0][1] + p[1] * m[1][1] + p[2] * m[2][1] + m[3][1],
          p[0] * m[0][2] + p[1] * m[1][2] + p[2] * m[2][2] + m[3][2]]

def point_distance(p1, p2):
  return math.sqrt((p2[0] - p1[0]) ** 2 + (p2[1] - p1[1]) ** 2 + (p2[2] - p1[2]) ** 2)

def vector_by_matrix(p, m):
  return [p[0] * m[0][0] + p[1] * m[1][0] + p[2] * m[2][0],
          p[0] * m[0][1] + p[1] * m[1][1] + p[2] * m[2][1],
          p[0] * m[0][2] + p[1] * m[1][2] + p[2] * m[2][2]]

def vector_length(v):
  return math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])

def vector_normalize(v):
  l = math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])
  return v[0] / l, v[1] / l, v[2] / l

def vector_dotproduct(v1, v2):
  return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]

def vector_crossproduct(v1, v2):
  return [
    v1[1] * v2[2] - v1[2] * v2[1],
    v1[2] * v2[0] - v1[0] * v2[2],
    v1[0] * v2[1] - v1[1] * v2[0],
    ]

def vector_angle(v1, v2):
  s = vector_length(v1) * vector_length(v2)
  f = vector_dotproduct(v1, v2) / s
  if f >=  1.0: return 0.0
  if f <= -1.0: return math.pi / 2.0
  return math.atan(-f / math.sqrt(1.0 - f * f)) + math.pi / 2.0

def blender_bone2matrix(head, tail, roll):
  # Convert bone rest state (defined by bone.head, bone.tail and bone.roll)
  # to a matrix (the more standard notation).
  # Taken from blenkernel/intern/armature.c in Blender source.
  # See also DNA_armature_types.h:47.
  target = [0.0, 1.0, 0.0]
  delta  = [tail[0] - head[0], tail[1] - head[1], tail[2] - head[2]]
  nor    = vector_normalize(delta)
  axis   = vector_crossproduct(target, nor)
  
  if vector_dotproduct(axis, axis) > 0.0000000000001:
    axis    = vector_normalize(axis)
    theta   = math.acos(vector_dotproduct(target, nor))
    bMatrix = matrix_rotate(axis, theta)
    
  else:
    if vector_crossproduct(target, nor) > 0.0: updown =  1.0
    else:                                      updown = -1.0
    
    # Quoted from Blender source : "I think this should work ..."
    bMatrix = [
      [updown, 0.0, 0.0, 0.0],
      [0.0, updown, 0.0, 0.0],
      [0.0, 0.0, 1.0, 0.0],
      [0.0, 0.0, 0.0, 1.0],
      ]
  
  rMatrix = matrix_rotate(nor, roll)
  return matrix_multiply(rMatrix, bMatrix)

#=====================================================================================
# GUI drawing and interaction


def GuessMeshNameList():
	"""Generate a list of object names for meshes to export.  Selected meshes are the
	first in the list."""
	MeshNameList = []
	# first put in selecte meshes
	Objects = Object.GetSelected()
	if len(Objects) > 0:
		for o in Objects:
			data = o.getData()
			if type(data) is Types.NMeshType:
				MeshNameList.append(o.getName())
	# the put in the rest of the meshes
	Objects = Object.Get()
	if len(Objects) > 0:
		for o in Objects:
			# skip if already in the list
			data = o.getData()
			if type(data) is Types.NMeshType:
				alreadythere = 0
				for k in MeshNameList:
					if k==o.getName():
						alreadythere =1
				if alreadythere == 0:
					MeshNameList.append(o.getName())
	MeshNameList.append('No Mesh')
	return MeshNameList

def GuessArmatureNameList():
	"""Generate a list of object names for meshes to export.  Selected meshes are the
	first in the list."""
	ArmatureNameList = []
	# first put in selecte meshes
	Objects = Object.GetSelected()
	if len(Objects) > 0:
		for o in Objects:
			data = o.getData()
			if type(data) is Types.ArmatureType:
				ArmatureNameList.append(o.getName())
	# the put in the rest of the meshes
	Objects = Object.Get()
	if len(Objects) > 0:
		for o in Objects:
			# skip if already in the list
			data = o.getData()
			if type(data) is Types.ArmatureType:
				alreadythere = 0
				for k in ArmatureNameList:
					if k==o.getName():
						alreadythere =1
				if alreadythere == 0:
					ArmatureNameList.append(o.getName())
	ArmatureNameList.append('No Armature')
	return ArmatureNameList

COMMAND_PROCESS = 1001
COMMAND_PATHCHANGE = 1002
COMMAND_MESHCHANGE = 1003
COMMAND_ARMATURECHANGE = 1004
COMMAND_CONTINUE = 1005

GUIMODE_CONFIGURE = 2001
GUIMODE_DIALOG = 2002

gMeshList = GuessMeshNameList()
gArmatureList = GuessArmatureNameList()
gMeshIndex = 0
gMeshMenu = Draw.Create("")
gArmatureIndex = 0
gArmatureMenu = Draw.Create("")
gOutputPathEdit = Draw.Create(os.path.join(characterdir, defaultfilename))
gGUIMode = GUIMODE_CONFIGURE
gGUIDialogString = "error: uninitialized dialog string"

def ExitGUI():
	Draw.Exit()

def DrawGUI():
	if gGUIMode == GUIMODE_CONFIGURE:
		DrawConfigureGUI()
	if gGUIMode == GUIMODE_DIALOG:
		DrawDialogGUI()

def DrawConfigureGUI():
	global gOutputPathEdit, gMeshIndex, gArmatureIndex, gMeshMenu, gArmatureMenu, gMeshList, gArmatureList
	BGL.glClearColor(0.2,0.2,0.2,0.0)
	BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
	BGL.glColor3f(1,1,1)
	BGL.glRasterPos2i(20,20)
	Draw.Text("Testing")
	BGL.glRasterPos2i(20,50)
	gOutputPathEdit = Draw.String("File path:", COMMAND_PATHCHANGE, 20,50,520,32,gOutputPathEdit.val,255, "Directory and base filename")
	Draw.Button("Generate", COMMAND_PROCESS, 20,90,100,32)
	MeshOptions = [ "%s %%x%d" % (gMeshList[x],x) for x in range(len(gMeshList))]
	ArmatureOptions = [ "%s %%x%d" % (gArmatureList[x],x) for x in range(len(gArmatureList))]
	gMeshMenu = Draw.Menu("Mesh name: %t|" + string.join(MeshOptions, "|"), COMMAND_MESHCHANGE, 20,130,200,32, gMeshIndex)
	gArmatureMenu = Draw.Menu("Armature name: %t|" + string.join(ArmatureOptions, "|"), COMMAND_ARMATURECHANGE, 20,160,200,32, gArmatureIndex)
	
def DrawDialogGUI():
	global gGUIMode, gGUIDialogString
	pass

def ProcessInputGUI(event, val):
	global gOutputPathEdit, gMeshIndex, gArmatureIndex, gMeshMenu, gArmatureMenu, gMeshList, gArmatureList
	if event == Draw.ESCKEY:
		ExitGUI()

def ProcessCommandGUI(command):
	global gOutputPathEdit, gMeshIndex, gArmatureIndex, gMeshMenu, gArmatureMenu, gMeshList, gArmatureList
	global gGUIMode, gGUIDialogString
	if command == COMMAND_PROCESS:
		processobjects(gMeshList[gMeshIndex], gArmatureList[gArmatureIndex], gOutputPathEdit.val )
	if command == COMMAND_MESHCHANGE:
		gMeshIndex = gMeshMenu.val
		#print "new mesh: ", gMeshIndex, gMeshList[gMeshIndex]
	if command == COMMAND_ARMATURECHANGE:
		gArmatureIndex = gArmatureMenu.val

def RegisterGUI():
	Draw.Register(DrawGUI, ProcessInputGUI, ProcessCommandGUI)


#=====================================================================================
# MESH/SKELETON/ANIMATION DATA PROCESSING

def GenActionList():
	"""Build a list of actions in the current blender scene"""
	ipos = Ipo.Get()
	actionlist = {}
	for i in ipos:
		(action, bone) = i.getName().split('.')
		actionlist[action] = 0
	return [x for x in actionlist.keys()]

class NebulaJoint:
	def __init__(self, name, index, parent, blenderbone, skeleton):
		"""Process the bone with the given name, index, parent joint, blender bone data structure,
		and a refernce to the NebulaSkeleton holding all this info."""
		self.children=[]
		self.name = name
		self.parent = parent
		self.parentindex = -1
		self.index = index
		self.bbone = blenderbone
		self.head = blenderbone.getHead()
		self.tail = blenderbone.getTail()
		self.skeleton = skeleton
		self.computeTransform()
		#print name, self.head, self.tail, self.rot, self.loc
		self.matrix = matrix_translate(quaternion2matrix(self.rot), self.loc)
		if parent:
			self.matrix = matrix_multiply(parent.matrix, self.matrix)
			self.parentindex = parent.index
			parent.children.append(self)

	def makeNebulascript(self):
		"""Generate the proper python script line to define this joint.  You'll need to
		prepend the python variable name referring to the node."""
		pos = self.loc
		quat = self.rot
		jointtext = "addjoint(%d, '%s', %d, %f,%f,%f, %f,%f,%f,%f)" % (
			self.index, self.name, self.parentindex,
			pos[0], pos[1], pos[2],
			quat[0], quat[1], quat[2], quat[3])
		return jointtext

	def computeTransform(self):
		"""Compute the translation and rotation from from the head->tail values.  The
		rotation is not expressed internally to blender as a quaternion, but instead is
		implied from the head->tail vector and a scalar roll value."""
		head = self.head
		tail = self.tail
		quat = matrix2quaternion(blender_bone2matrix(head,tail, self.bbone.getRoll()))
		if self.parent:
			# Compute the translation from the parent bone's head to the child
			# bone's head, in the parent bone coordinate system.
			# The translation is parent_tail - parent_head + child_head,
			# but parent_tail and parent_head must be converted from the parent's parent
			# system coordinate into the parent system coordinate.
			parent_invert_transform = matrix_invert(quaternion2matrix(self.parent.rot))
			parent_head = vector_by_matrix(self.parent.head, parent_invert_transform)
			parent_tail = vector_by_matrix(self.parent.tail, parent_invert_transform)
			self.rot = quat
			self.loc = [parent_tail[0] - parent_head[0] + head[0], parent_tail[1] - parent_head[1] + head[1], parent_tail[2] - parent_head[2] + head[2]]
		else:
			# Apply the armature's matrix to the root bones
			matrix = self.skeleton.matrix
			head = point_by_matrix(head, matrix)
			tail = point_by_matrix(tail, matrix)
			self.rot = matrix2quaternion(matrix_multiply(matrix, quaternion2matrix(quat))) # Probably not optimal
			self.loc = head
			self.head = head
			self.tail = tail

class NebulaSkeletonGenerator:
	def __init__(self, armatureobject):
		armature = armatureobject.getData()
		self.jointlist = []
		self.matrix = armatureobject.getMatrix()
		self.computeBoneMapping(armature)
		self.generateJointList(armature)

	def computeBoneMapping(self, armature):
		"""Given a blender armature, build up an internal mapping to
		map from a bone name to the bone index."""
		self.bonemapping = {}
		for b in armature.getBones():
			self.addBoneMap(b)

	def addBoneMap(self, bone):
		bonename = bone.getName()
		# add virtual bone if needed
		bonehead = bone.getHead()
		boneindex = len(self.bonemapping.keys())
		self.bonemapping[bonename] = boneindex
		for b in bone.getChildren():
			self.addBoneMap(b)

	def generateJointList(self, skeleton):
		for b in skeleton.getBones():
			self.addBoneTree(b)

	def addBoneTree(self, rootbone):
		self.addBone(rootbone)
		for b in rootbone.getChildren():
			self.addBoneTree(b)

	def addBone(self, bone):
		jointname = bone.getName()
		jointindex = self.bonemapping[jointname]
		parent = None
		if (bone.hasParent()):
			parentindex = self.bonemapping[bone.parent.getName()]
			parent = self.jointlist[parentindex]
		newjoint = NebulaJoint(jointname, jointindex, parent, bone, self)
 		self.jointlist.append(newjoint)
	
	def writeNebulaScript(self, outstream):
		charnodename = 'char'
		self.writeNebulaPrelude(charnodename, outstream)
		outstream.write("%s.beginjoints(%d)\n" % (charnodename, len(self.jointlist)) )
		for j in self.jointlist:
			outstream.write("%s.%s\n" % (charnodename, j.makeNebulascript()) )
		outstream.write("%s.endjoints()\n" % charnodename)
		self.writeNebulaEpilogue(charnodename, outstream)
		return len(self.jointlist)

	def writeNebulaPrelude(self, charnodename, outstream):
		outstream.write("%s = new('ncharacternode','%s')\n" % (charnodename, charnodename))
		outstream.write("%s.setchannel('time')\n" % charnodename)
		outstream.write("%s.setscale(1.0)\n" % charnodename)
		outstream.write("%s.setanimfile('character.anim')\n" % charnodename)
		outstream.write("%s.setstatechannel('actor')\n" % charnodename)
		outstream.write("%s.setreadonly(0)\n" % charnodename)
		outstream.write("%s.setskinmesh('skin')\n" % charnodename)
		outstream.write("%s.setcastshadow(0)\n" % charnodename)
		outstream.write("%s.setrenderskeleton(1)\n" % charnodename)
		outstream.write("%s.setreptype('loop')\n" % charnodename)
		outstream.write("sel('%s')\n" % charnodename)
		outstream.write("skin = new('nmeshnode','skin')\n")
		outstream.write("skin.setfilename('character.n3d')\n")
		outstream.write("skin.setreadonly(1)\n")
		outstream.write("skin.setcastshadow(0)\n")
		

	def writeNebulaEpilogue(self, charnodename, outstream):
		actionlist = GenActionList()
		numactions = len(actionlist)
		outstream.write( "%s.beginstates(%d)\n" % (charnodename, numactions) )
		for s in range(numactions):
			actionname = actionlist[s]
			outstream.write( "%s.addstate(%d,'%sS')\n" % (charnodename, s, actionname) )
			outstream.write( "%s.beginstateanims(%d,1)\n" % (charnodename, s) )
			outstream.write( "%s.addstateanim(%d, 0, '%s', '%s')\n" % (charnodename, s, actionname, actionname) )
			outstream.write( "%s.endstateanims(%d)\n" % (charnodename, s) )
		outstream.write( "%s.endstates()\n" % charnodename )
		outstream.write( "sel('..')\n" )
		outstream.write( """shd = new('nshadernode','sn')
shd.setnumstages(1)
shd.setcolorop(0,'mul tex prev')
shd.begintunit(0)
shd.setaddress('wrap','wrap')
shd.setminmagfilter('linear','linear')
shd.settexcoordsrc('uv0')
shd.setenabletransform(0)
shd.endtunit()
shd.setlightenable(1)
shd.setdiffuse(1,1,1,1)
shd.setemissive(0,0,0,0)
shd.setambient(1,1,1,1)
shd.setalphaenable(0)
tex = new('ntexarraynode','tex')
tex.settexture(0,'character.bmp','none')
""" )

# find all the weights for vertices in the mesh

class NebulaMeshGenerator:
	"""Class to process the blender mesh and armature, to generate a proper
	nebula .n3d file, flattened w/ proper uv coordinates and joint weights."""

	def __init__(self, blendermeshobject, nebulaskeleton = None):
		self.objmatrix = blendermeshobject.getMatrix()
		self.ProcessMesh(blendermeshobject.getData(), nebulaskeleton)

	def ProcessMesh(self, blendermesh, nebulaskeleton):
		"""Compute the vertex, face, and joint data for the given mesh and armature."""
		self.bmesh = blendermesh
		self.nskeleton = nebulaskeleton
		self.vertices = []
		self.ProcessFaces()
		self.jointweights = None
		if (self.nskeleton):
			self.ComputeJointWeights()

	def GetVertex(self, vcindex, uvcoords):
		"""Get vertex index for the given coordinate index/uv index pair.  Generates a new vertex
		if needed."""
		# match with a current vertex?
		for v in self.vertices:
			if ( (v['vcindex'] == vcindex) and (v['uv0'] == uvcoords[0]) and (v['uv1'] == uvcoords[1]) ):
				return v['nebindex']
		# add a new one
		newv = {'nebindex':len(self.vertices)+1, 'vcindex':vcindex, 'uv0':uvcoords[0], 'uv1':uvcoords[1] }
		self.vertices.append(newv)
		#print newv
		return newv['nebindex']
		
	def ProcessFaces(self):
		"""Generate a list of faces, where each face is a triangle consisting of vertex indices.  The
		mesh is flattened, meaning every the vertex indices for coordinate, uv coordinate, and normal are
		all equal, of the form n/n/n in the mesh file."""
		self.faces = []
		for f in self.bmesh.faces:
			# triangularize if needed
			for triangleindex in range(len(f.v)-2):
				newface = []
				vindices = [0, triangleindex+1, triangleindex+2]
				for vindex in vindices:
					findex = vindex
					thisvertexindex = f.v[findex].index
					if (f.uv):
						thisuvcoord = f.uv[findex]
					else: thisuvcoord = (0,0)
					nebindex = self.GetVertex(thisvertexindex, thisuvcoord)
					newface.append(nebindex)
				#print newface
				self.faces.append(newface)

	def ComputeJointWeights(self):
		"""Go through the vertex list and compute the joint indices and weights."""
		self.jointweights = []
		bonemap = self.nskeleton.bonemapping
		for f in self.vertices:
			vcindex = f['vcindex']
			blendweights = self.bmesh.getVertexInfluences(vcindex)
			# map from bone name to joint index using the nebulaskeleton we were given
			#print blendweights
			# blender weights do not have to add to one, so we must normalize
			weightsum = reduce(operator.add, [x[1] for x in blendweights])
			# now map from bone name to joint index, and normalize weights too
			nebweights = [ (bonemap[x[0]], x[1]/weightsum) for x in blendweights ]
			#print nebweights
			self.jointweights.append(nebweights)

	def writeNebulaMesh(self, outfile):
		"""After processing the mesh and joints, this function will write out a nebula mesh to
		the given file or I/O object. Returns a tuple (v,f) with the number of vertices and faces
		written"""
		self.writeNebulaPrelude(outfile)
		self.writeNebulaVertices(outfile)
		self.writeNebulaFaces(outfile)
		self.writeNebulaEpilogue(outfile)
		return ( len(self.vertices), len(self.faces) )

	def writeNebulaPrelude(self, outfile):
		pass

	def writeNebulaEpilogue(self, outfile):
		pass

	def writeNebulaVertices(self, outfile):
		if (self.jointweights):
			if (len(self.jointweights) <> len(self.vertices)):
				print "Warning: joint weights do not match up with vertices"
				dojoints = 0
			else:
				dojoints = 1
		else:
			dojoints = 0
		for vindex in range(len(self.vertices)):
			# find the blender vertex that holds most of the data we want
			v = self.vertices[vindex]
			bv = self.bmesh.verts[v['vcindex']]
			vcoords = point_by_matrix(bv.co, self.objmatrix)
			vnormal = vector_by_matrix(bv.no, self.objmatrix)
			outfile.write("v %f %f %f\n" % tuple(vcoords) )
			outfile.write("vn %f %f %f\n" % tuple(vnormal) )
			outfile.write("vt %f %f\n" % (v['uv0'], v['uv1']) )
			if (dojoints):
				jointinfo = self.jointweights[vindex]
				if (len(jointinfo) > 4):
					print "Error: a vertex has > 4 joints influencing it, using only the first four."
					jointinfo = jointinfo[0:4]
				jointbits = ["%d %f" % x for x in jointinfo]
				jointheader = "jw%d " % len(jointinfo)
				jointline = jointheader + string.join(jointbits) + "\n"
				outfile.write(jointline)
				

	def writeNebulaFaces(self, outfile):
		for f in self.faces:
			facebits = ["%d/%d/%d" % (x,x,x) for x in f]
			faceline = "f %s\n" % string.join(facebits)
			outfile.write(faceline)
				


def CalcIpoName(ipo):
	"""Hack we need to use until they put in the
	quaternion ipo names.  It turns out we can
	deduce the ipo name from the repr string:
	repr(ipo) ="IpoCurve xxxxx ### xxx xxx ..."
	Where "###" is a number from 25-28, specifying
	the curve name:
	25=QuatW, 26=QuatX, 27=QuatY, 28=QuatZ"""
	#curverep = repr(ipo)
	#print dir(ipo), curverep,
	#curvebits = curverep.split(' ')
	#curvevalue = int(curvebits[2])
	#curvename = 'Empty'
	# if it's not a quaternion, catch the exception and fallback to 
	# normal getName() functionality
	#try:
	#	curvename = ('QuatW','QuatX','QuatY','QuatZ')[curvevalue-25]
	#except:
	#	curvename = ipo.getName()	
	#	print "Error: ipo curve %s is not part of a quaternion" % curvename
	curvename = ipo.getName()
	#print "ipocurve ", curvename
	return curvename

class NAnimCurve:
	"""Seeks to represent the essential elements of a Nebula1 NAnimCurve"""

	def __init__(self):
		self.Reset()

	def Reset(self):
		self.SetKeysPerSecond(12)
		self.SetStartKey(0)
		self.SetInterpolationType('linear')
		self.SetRepeatType('oneshot')
		self.keys = []
		
	def AddKey(self, key):
		self.keys.append(key)

	def SetKeysPerSecond(self, kps):
		self.kps = kps

	def SetStartKey(self, startkey):
		self.startkey = startkey

	def SetInterpolationType(self, ipoltype):
		"""Set interpolation between keys to 'step','linear', or 'quaternion'"""
		self.ipoltype = ipoltype

	def SetRepeatType(self, repeattype):
		"""Set animation repeat to be 'loop' or 'oneshot'"""
		self.repeattype = repeattype

	def DumpAnimText(self, statename, actionname, bonename, iostream):
		"""Given the state, action, and bone name, dump out text data for
		use within a Nebula1 .anim file"""
		curvename = "%s_%s_%s_%s" % (statename, actionname, 'rot', bonename)
		declarationline = "curve %s %d %d %f %s %s\n" % (curvename, self.startkey, len(self.keys),
			self.kps, self.repeattype, self.ipoltype)
		keylines = [ ("key %f %f %f %f\n" % tuple(x)) for x in self.keys ]
		iostream.writelines( [declarationline] + keylines )

class ActionBundle:
	"""Class which holds all the actions and their
	associated curves.  The class will run through all the
	ipos and build up a set of actions.  Each action, with a
	unique name, will hold a list of its curves.  To use this
	class, you need to build your animated models such that
	the ipos for each action are of the form "Action.Bone" where
	the names exactly match the relevant Action and Bone.
	Then you can just invoke."""

	def __init__(self, nebskeleton):
		self.actions = {}
		self.nebskeleton = nebskeleton

	def ProcessIpo(self,ipo):
		"""Add the curves of the ipo to the correct action"""
		iponame = ipo.getName()
		print iponame
		(actionname, bonename) = iponame.split('.')
		#print "action %s, bone %s" % (actionname, bonename)
		# add the curve to the list of curves for this action
		blenderboneindex = self.nebskeleton.bonemapping[bonename]
		blenderbone = self.nebskeleton.jointlist[blenderboneindex]
		thisanimcurve = self.MakeAnimCurve(ipo, blenderbone)
		try:
			self.actions[actionname][bonename] = thisanimcurve
		except KeyError:
			self.actions[actionname] = {}
			self.actions[actionname][bonename] = thisanimcurve

	def MakeAnimCurve(self, ipo, blenderbone):
		"""Given a Blender ipo, generate an animcurve, corresponding to data in the 
		Nebula1 AnimCurve data structure."""
		newcurve = NAnimCurve()
		newcurve.SetInterpolationType('quaternion')
		newcurve.SetRepeatType('loop')
		curvepoints = [ x.getPoints() for x in ipo.getCurves() ]
		# figure out how many keys, and keys/second.  Assume 50 frames/sec
		basecurve = curvepoints[0]
		starttime = 0.0
		endtime = basecurve[-1].getPoints()[0]
		keyspacing = 1 # self.FindIpoMinSpacing(ipo)
		keyrate = 50.0 / keyspacing
		numkeys = max( 1, (endtime - starttime) / keyspacing)
		newcurve.SetKeysPerSecond(keyrate)
		keytimes = [x*keyspacing for x in range(numkeys)]
		numcurves = ipo.getNcurves()
		# build an array to map from blender curve index to nebula index
		curvemap = self.BuildCurveMap(ipo)
		desiredcurveorder = ('QuatX','QuatY','QuatZ','QuatW')
		for ktime in keytimes:
			# construct the key for this time, then swizzle the values into the right order
			curkeyvalues = [ ipo.EvaluateCurveOn(i, ktime) for i in range(numcurves) ]
			newkey = [ curkeyvalues[curvemap[x]] for x in desiredcurveorder ]
			# multiply the quat by the bone's quaternion to get a final value
			newkey = quaternion_multiply(newkey, blenderbone.rot)
			newcurve.AddKey(newkey)
		return newcurve

	def BuildCurveMap(self, ipo):
		"""Given a blender ipo, build a list of indices to map from
		the blender curve indices to the nebula indices.  This is needed
		since the two curve orders may not match. For example, nebula
		wants the curves for a quaternion in (x,y,z,w) order, whereas
		in blender they could be (w,x,y,z).  We have to examine the
		blender curve labels to be sure."""
		curvemap = {}
		ipocurves = ipo.getCurves()
		for i in range(ipo.getNcurves()):
			thiscurveid = CalcIpoName(ipocurves[i])
			curvemap[thiscurveid] = i
		return curvemap		

	def FindCurveMinSpacing(self, ipocurve):
		"""Given an ipo curve, figures out the smallest spacing
		between any two keys"""
		keytimes = [ x.getPoints()[0] for x in ipocurve.getPoints() ]
		#print keytimes
		keyspacings = [keytimes[i+1] - keytimes[i] for i in range(len(keytimes)-1) ]
		if len(keyspacings) < 1:
			return 1.0
		spacing = min(keyspacings)
		if spacing < 1:
			return 1.0
		else:
			return spacing

	def FindIpoMinSpacing(self, ipo):
		"""Given an ipo, figure out the smallest spacing of any of the
		ipo curves.  Sometimes an ipo will  have multiple curves with
		different key times and spacings; we should probably take the
		smallest one"""
		curvespacings = [ self.FindCurveMinSpacing(x) for x in ipo.getCurves() ]
		return min(curvespacings)

	def DumpCurves(self, file):
		"""Print out all the current curves that have been computed so far"""
		for actionname in self.actions.keys():
			statename = actionname + 'S'
			for curve in self.actions[actionname].items():
				(bonename, animcurve) = curve
				animcurve.DumpAnimText(statename, actionname, bonename, file)

	def GenAnimText(self,file):
		"""Output to file the text of a proper nebula .nanim file"""
		file.write("curves %d\n" % self.CalcTotalCurves())
		self.DumpCurves(file)

	def CalcTotalCurves(self):
		"""Calculate how many curves we have total in the bundle."""
		countperaction = [ len(x[1]) for x in self.actions.items() ]			
		totalcount = reduce(operator.add, countperaction)
		return totalcount

def processobjects(meshname, armaturename, characterfilebase):
	print "armature %s, mesh %s" % (armaturename, meshname)

	if (armaturename == 'No Armature'):
		armatureobject = None
	else:
		armatureobject = Object.Get(armaturename)
	if (meshname == 'No Mesh'):
		meshobject = None
	else:
		meshobject = Object.Get(meshname)

	meshfilename = characterfilebase + ".n3d"
	scriptfilename = characterfilebase + ".py"
	animfilename = characterfilebase + ".anim"

	meshfile = file(meshfilename,'w')
	scriptfile = file(scriptfilename,'w')
	animfile = file(animfilename, 'w')

	# process the armature first, since we need to know the joint/index mapping to
	# correctly generate joint weights for the .obj mesh file
	if (armatureobject):
		nebskeleton = NebulaSkeletonGenerator(armatureobject)
		nebskeleton.writeNebulaScript(scriptfile)
	else:
		nebskeleton = None	

	# process the mesh object
	print meshobject
	if (meshobject):
		print "processing mesh"
		nebmesher = NebulaMeshGenerator(meshobject, nebskeleton)
		nebmesher.writeNebulaMesh(meshfile)

	# process the animation ipos
	if (nebskeleton):
		actions = ActionBundle(nebskeleton)
		curves = Ipo.Get()
		for cipo in curves:
			actions.ProcessIpo(cipo)
		actions.GenAnimText(animfile)

RegisterGUI()